/*
 * Copyright (C) 2010 Mark Hills <mark@pogo.org.uk>
 *
 * Listen for an asynchronous event from a PostgreSQL server, and call
 * the given command when it happens.
 */

#include <assert.h>
#include <errno.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>
#include <syslog.h>
#include <string.h> /* strsignal() */
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "libpq-fe.h"

void lerror(const char *prefix)
{
	syslog(LOG_ERR, "%s: %s", prefix, strerror(errno));
}

int fork_exec(const char *path, char *const argv[])
{
	int status;
	pid_t pid;

	pid = fork();

	if (pid == -1) {
		lerror("fork");
		return -1;
	}

	if (pid == 0) { /* child */
		execvp(path, argv);
		lerror("execvp");
		exit(EX_OSERR);
	}

	if (waitpid(pid, &status, 0) == -1) {
		lerror("waitpid");
		return -1;
	}

	if (WIFEXITED(status) && WEXITSTATUS(status) == 0)
		return 0;

	if (WIFSIGNALED(status)) {
		syslog(LOG_ERR, "Child process terminated by signal %s",
			strsignal(WTERMSIG(status)));
	}

	if (WIFEXITED(status) && WEXITSTATUS(status) != 0) {
		syslog(LOG_ERR, "Child process exited with return code %d",
			WEXITSTATUS(status));
	}

	syslog(LOG_ERR, "Child process exited abnormally");
	return -1;
}

void usage(FILE *fd)
{
	fprintf(fd, "Usage: pgnotifyd [-br] [-d <conninfo>] <event> [--] <command> [ <arg> ... ]\n");
}

int main(int argc, char *argv[])
{
	int sd, opt;
	const char *event;
	char *sql, *cs = "";
	bool bg = false, on_start = true;
	fd_set input_mask;
	PGconn *conn;
	PGresult *res;
	PGnotify *notify;

	for (;;) {
		opt = getopt(argc, argv, "bd:hr");
		if (opt == -1)
			break;

		switch (opt) {
		case 'd':
			cs = optarg;
			break;
		case 'b':
			bg = true;
			break;
		case 'h':
			usage(stdout);
			return EX_OK;
		case 'r':
			on_start = false;
			break;
		default:
			usage(stderr);
			return EX_USAGE;
		}
	}

	argv += optind;
	argc -= optind;

	if (argc < 1) {
		usage(stderr);
		return EX_USAGE;
	}

	event = argv[0];

	/* Syslog */

	openlog("pgnotifyd", LOG_PID | LOG_PERROR, LOG_DAEMON);

	/* Make the connection */

	conn = PQconnectdb(cs);
	if (PQstatus(conn) != CONNECTION_OK) {
		syslog(LOG_ERR, "%s", PQerrorMessage(conn));
		return EX_UNAVAILABLE;
	}

	/* Send the initial LISTEN command */

	(void)asprintf(&sql, "LISTEN %s", event);
	res = PQexec(conn, sql);
	if (PQresultStatus(res) != PGRES_COMMAND_OK) {
		syslog(LOG_ERR, "%s", PQerrorMessage(conn));
		PQclear(res);
		return EX_UNAVAILABLE;
	}
	PQclear(res);

	/* Prepare the data needed for select() */

	sd = PQsocket(conn);
	assert(sd != -1);

	FD_ZERO(&input_mask);
	FD_SET(sd, &input_mask);

	/* Go to the background if required */

	if (bg) {
		if (daemon(0, 0) == -1) {
			lerror("daemon");
			return EX_OSERR;
		}
	}

	/* Catch database changes where we missed the notifications */

	if (on_start) {
		if (fork_exec(argv[1], &argv[1]) == -1)
			return EX_OSERR;
	}

	/* The main loop */

	for (;;) {
		unsigned int nevents = 0;

		if (select(sd + 1, &input_mask, NULL, NULL, NULL) == -1) {
			lerror("select");
			return EX_OSERR;
		}

		if (PQconsumeInput(conn) == 0) {
			syslog(LOG_ERR, "%s", PQerrorMessage(conn));
			return EX_UNAVAILABLE;
		}

		while ((notify = PQnotifies(conn)) != NULL) {
			PQfreemem(notify);
			nevents++;
		}

		if (nevents > 0) {
			if (fork_exec(argv[1], &argv[1]) == -1)
				return EX_OSERR;
		}
	}

	PQfinish(conn);
	closelog();

	return EX_OK;
}
